#!/usr/bin/env python3

"""
Controls the training and prediction steps for TIdeS.

Dependencies include: Optuna, Scikit-learn.
"""

import pickle
import random
import sys
import time
from collections import defaultdict
from datetime import timedelta
from pathlib import Path

import optuna
import numpy as np

from sklearn.dummy import DummyClassifier
from sklearn.svm import SVC
from sklearn.model_selection import train_test_split, cross_val_score
from sklearn.metrics import classification_report

def objective_svm(trial, X_features, X_labels, threads):
    train_x, test_x, train_y, test_y = train_test_split(
                                        X_features,
                                        X_labels,
                                        shuffle = True,
                                        test_size = 0.25,
                                        random_state = random.randint(0,10000),
                                        stratify = X_labels
                                        )
    params = {
        'C': trial.suggest_float('C', 1e-8, 10),
        'kernel': trial.suggest_categorical('kernel', ['linear', 'rbf']),
        'random_state': random.randint(0,10000),
        }

    clf = SVC(**params)
    clf_score = cross_val_score(
                    clf,
                    train_x,
                    train_y,
                    n_jobs = int(threads),
                    cv = 5
                    ).mean()

    return clf_score


def train_model(train_data, threads, contam):
    X_features, X_labels = train_data[1:]
    optuna.logging.set_verbosity(optuna.logging.WARNING)

    # optuna.logging.set_verbosity(optuna.logging.INFO)

    study = optuna.create_study(direction='maximize')

    study.optimize(lambda trial: objective_svm(trial, X_features, X_labels, threads), n_jobs = int(threads), n_trials = 100)

    opt_trial = study.best_trial

    opt_clf = SVC(probability = True, random_state = random.randint(0,10000), **opt_trial.params)

    dm_clf = DummyClassifier(strategy = "stratified", random_state = random.randint(0,10000))

    train_x, test_x, train_y, test_y = train_test_split(
                                        X_features,
                                        X_labels,
                                        shuffle = True,
                                        test_size = 0.25,
                                        random_state = random.randint(0,10000),
                                        stratify = X_labels
                                        )

    opt_clf.fit(train_x, train_y)
    dm_clf.fit(train_x, train_y)

    opt_preds = opt_clf.predict(test_x)
    dm_preds = dm_clf.predict(test_x)

    opt_clf_report = classification_report(test_y, opt_preds)
    dm_clf_report = classification_report(test_y, dm_preds)

    return opt_clf, study, (opt_clf_report, dm_clf_report)


def parse_qpreds(q_names, q_preds, eval_names, contam = False):
    summary = defaultdict(list)

    for n in range(0, len(q_preds)):
        if contam:
            summary[q_names[n]] += [f'{eval_names[np.argmax(q_preds[n])]}']
            for i in q_preds[n]:
                summary[q_names[n]] += [f'{i:.3f}']

        else:
            summary[q_names[n].split('.pORF')[0]].append([
                                                    f'{q_names[n].split()[0]}',
                                                    f'{q_names[n]}',
                                                    f'{eval_names[np.argmax(q_preds[n])]}',
                                                    f'{q_preds[n][0]:.4f}',
                                                    f'{q_preds[n][1]:.4f}'
                                                    ])
    return summary


def distance(co1, co2):
    # Capture distance between two sets of points
    return ((abs(co1[0] - co2[0])**2) + (abs(co1[1] - co2[1])**2))**0.5


def best_preds(pred_list):
    preds_to_keep = []
    crf_evals = sorted([i for i in pred_list if 'CRF' in i], key = lambda x: (-float(x[-1]), -int(x[1].split()[2].split(':')[-1])))

    if crf_evals:

        preds_to_keep = [crf_evals[0]]
        coords_eval = [tuple(sorted(int(i) for i in crf_evals[0][1].split(':')[-1].split('(')[0].split('-')))]

        for pred in crf_evals[1:]:
            crds = sorted(int(i) for i in pred[1].split(':')[-1].split('(')[0].split('-'))
            nearest = min(coords_eval, key = lambda x: distance(x, crds))

            if crds[0] <= nearest[1] and crds[1] >= nearest[0]:
                continue

            else:
                preds_to_keep.append(pred)
                coords_eval.append(tuple(crds))

    return preds_to_keep


def classify_orfs(taxon_code: str, start_time, train_data, query_data, threads: int = -1, model = None, contam = False, verb = True):

    clf_dir = f'{taxon_code}_TIdeS/Classification/'
    clf_tsv = f'{clf_dir}{taxon_code}.TIdeS_Classification.tsv'
    clf_tp_sb_tsv = f'{clf_dir}{taxon_code}.TIdeS_Classification.SingleBest.tsv'
    clf_scr_tsv = f'{clf_dir}{taxon_code}.TIdeS_Classification_Report.txt'
    clf_stdy = f'{clf_dir}{taxon_code}.TIdeS.OptunaStudy.pkl'

    non_contam_header = 'Transcript\tpORF\tFull-pORF-Name\tEvaluation\tProb-OOF\tProb-CRF\n'

    Path(clf_dir).mkdir(exist_ok = True, parents = True)

    query_names, query_features = query_data

    if model:
        trained_clf = model

    else:
        if verb:
            print(f'[{timedelta(seconds=round(time.time()-start_time))}]  Training TIdeS')

        trained_clf, opt_study, clf_scrs = train_model(train_data, threads, contam)

        with open(clf_scr_tsv, 'w+') as w:
            w.write('TIdeS-Classification-Report\n')
            w.write(f'{clf_scrs[0]}\n')
            w.write('"Dummy"-Classification-Report\n')
            w.write(f'{clf_scrs[1]}')

    if verb:
        print(f'[{timedelta(seconds=round(time.time()-start_time))}]  Classifying ORFs')

    trained_clf_qpreds = trained_clf.predict_proba(query_data[1])


    if not contam:
        eval_names = ['OOF', 'CRF']
    else:

        eval_names = trained_clf.classes_
        tmp = '\t'.join([f'Prob-{name}' for name in eval_names])
        contam_header = f'ORF\tEvaluation\t{tmp}\n'

    query_summary = parse_qpreds(query_data[0], trained_clf_qpreds, eval_names, contam)

    with open(clf_tsv, 'w+') as w:
        if contam:
            w.write(contam_header)
            for k, v in query_summary.items():
                x = '\t'.join(v)
                w.write(f'{k}\t{x}\n')

        else:
            w.write(non_contam_header)
            for k, v in query_summary.items():
                for i in v:
                    x = '\t'.join(i)
                    w.write(f'{k}\t{x}\n')
    if not model:
        with open(clf_stdy, 'wb') as fout:
            pickle.dump(opt_study, fout)

    if not contam:
        query_best_preds = {}
        for k, v in query_summary.items():
            bps = best_preds(v)
            if bps:
                query_best_preds[k] = bps

        with open(clf_tp_sb_tsv, 'w+') as w:
            w.write(non_contam_header)
            for k, v in query_best_preds.items():
                if v:
                    x = '\t'.join(v[0])
                    w.write(f'{k}\t{x}\n')

        return (query_summary, query_best_preds), trained_clf

    return query_summary, trained_clf



#!/usr/bin/env python3

"""
Extracts putative ORFs from a given transcriptome's FASTA-formatted file,
using a user-supplied genetic code and which meet a minimum length criterion.

Returns the putative ORFs and their sequences from all transcripts.


Dependencies include: BioPython, NumPy.
"""

import os, re, subprocess, sys, time

from datetime import timedelta
from pathlib import Path

import numpy as np

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord


def gcode_start_stops(gcode: str) -> list:
    """
    Provides a list of stop codons from a given genetic code (i.e. translation table).

    Parameters
    ----------
    gcode:  NCBI numerical genetic code

    Returns list of valid stop codons on plus and minus strands as well as translation table
    """

    ncbi_link = 'https://www.ncbi.nlm.nih.gov/Taxonomy/Utils/wprintgc.cgi'

    stop_codon_by_gcodes = {
        '1': [['TAA','TAG','TGA'],['TTA','CTA','TCA'], 1],
        '4': [['TAA','TAG'],['TTA','CTA'], 4],
        '6': [['TGA'],['TCA'], 6],
        '10': [['TAA','TAG'],['TTA','CTA'], 10],
        '12': [['TAA','TAG','TGA'],['TTA','CTA','TCA'], 12],
        '26': [['TAA','TAG','TGA'],['TTA','CTA','TCA'], 26],
        '29': [['TGA'],['TCA'], 29],
        '30': [['TGA'],['TCA'], 30],
        'universal': [['TAA','TAG','TGA'],['TTA','CTA','TCA'], 1],
        'blepharisma': [['TAA','TAG'],['TTA','CTA'], 4],
        'ciliate': [['TGA'],['TCA'], 6],
        'euplotes': [['TAA','TAG'],['TTA','CTA'], 10],
        'alt-yeast': [['TAA','TAG','TGA'],['TTA','CTA','TCA'], 12],
        'pachysolen': [['TAA','TAG','TGA'],['TTA','CTA','TCA'], 26],
        'mesodinium': [['TGA'],['TCA'], 29],
        'peritrich': [['TGA'],['TCA'], 30]
        }

    if str(gcode).lower() not in stop_codon_by_gcodes.keys():
        print('\nCheck the chosen genetic code/translation table.\n\n' \
            'Supported tables include:\n    ' +
            ',\n    '.join(list(stop_codon_by_gcodes.keys())))

        print(f'\nFor more information, visit: {ncbi_link}\n')

        return None

    return stop_codon_by_gcodes[str(gcode)]


def bin_pos(cdn_pos: list, rf: dict, strand: str = '+') -> dict:
    """
    Organizes all codon positions of interest by reading frame and strand.

    Parameters
    ----------
    cdn_pos:  list of codon positions
    rf:       reading-frame dictionary to populate
    strand:   plus or minus strand

    Returns
    ----------
    rf: dictionary of all codon positions by reading frame
    """

    rf_adj = 0
    if strand != '+':
        rf_adj += 3

    for i in cdn_pos:
        if i % 3 == 0:
            rf[1 + rf_adj].append(i)
        elif i % 3 == 1:
            rf[2 + rf_adj].append(i)
        else:
            rf[3 + rf_adj].append(i)

    return rf


def capture_start_stop_pos(seq: str, fstop: list, rstop: list) -> dict:
    """
    Identifies all start and stop codon position for a given sequence in all
    reading frames.

    Parameters
    ----------
    seq:    single transcript/sequence
    fstop:  list of stop codons in the plus strand orientation
    rstop:  list of stop codons in the minus strand orientation

    Returns
    ----------
    start_rf:  dictionary of start codon positions by reading frame
    stop_rf:   dictionary of stop codon positions by reading frame
    """

    stop_cdns_p, stop_cdns_m = [], []

    # Capture indices of start codons from both strands
    start_cdns_p = [m.start() for m in re.finditer('ATG', seq)]
    start_cdns_m = [m.end() for m in re.finditer('CAT', seq)]

    # Capture indices of all stop codons from both strands
    for n in range(len(fstop)):
        stop_cdns_p += [m.end() for m in re.finditer(fstop[n], seq)]
        stop_cdns_m += [m.start() for m in re.finditer(rstop[n], seq)]

    start_rf = bin_pos(start_cdns_p, {1:[], 2:[], 3:[], 4:[], 5:[], 6:[]})
    start_rf = bin_pos(start_cdns_m, start_rf, '-')

    stop_rf = bin_pos(stop_cdns_p, {1:[], 2:[], 3:[], 4:[], 5:[], 6:[]})
    stop_rf = bin_pos(stop_cdns_m, stop_rf, '-')

    return start_rf, stop_rf


def check_start_dist(tmp_start: list, stop_coord: int, rf: int) -> list:
    """
    Assesses putative start codon positions, keeping a set of start codons at
    least 15nt apart from one another. Note that this starts with the first
    codon (by position in array) in the "tmp_start" array of start codons positions.

    Parameters
    ----------
    tmp_start:   numpy array of start codon positions
    stop_coord:  stop codon coordinate for this putative ORF
    rf:          current reading frame

    Returns list of pORF coordinates for a single transcript fragment between stop codons
    """

    stp_pos = stop_coord[1]
    if rf > 3:
        stp_pos = stop_coord[0]

    # Grab the initial start codon coordinate and stop codon for this pORF
    porf_coords = [(tmp_start[0], stp_pos)]

    # Start tracking the last valid start codon
    last_valid = tmp_start[0]
    tmp_start = tmp_start[1:]

    # Track whether to skip the next start codon coordinate in the array
    skip_pos = False
    for n in range(len(tmp_start)):
        if skip_pos:
            skip_pos = False
            continue

        # Check that current start coordinate is at least 15nt from previous valid start codon
        elif abs(tmp_start[n] - last_valid) > 14:
            porf_coords.append((tmp_start[n], stp_pos))
            last_valid = tmp_start[n]

            # If final coordinate, no further comparison needed
            if n != len(tmp_start) - 1:

                # Evaluate skipping next codon
                if abs(tmp_start[n] - tmp_start[n+1]) < 14:
                    skip_pos = True
                else:
                    skip_pos = False
        else:
            pass

    return list(set(porf_coords))


def check_partial_pORFs(start_coords: np.ndarray, stop_coords: np.ndarray, rf: int, min_len: int, seq_len: int) -> list:
    """
    Checks for partial pORFs if no start codons could create a complete pORF
    (5' partial ORF), or no stop codon could create a complete pORF (3' partial ORF).

    Parameters
    ----------
    start_coords:  numpy array of start codon positions
    stop_coords:   numpy array of stop codon positions
    rf:            current reading frame
    min_len:       minimum ORF length to consider [default is 300nt]

    Returns
    ----------
     partial_coords:  list of partial pORF coordinates in the given reading frame
    """

    partial_coords = []
    if rf > 3:
        seq_start = rf - 4

    else:
        seq_start = rf - 1

    seq_end = seq_len - (seq_len - seq_start)%3

    if stop_coords.size == start_coords.size == 0:
        partial_coords.append((seq_start, seq_end))

    # Captures partial pORFs lacking a start codon in the current reading frame
    elif stop_coords.size != start_coords.size == 0:
        if rf > 3:
            if (seq_end - stop_coords[-1]) >= min_len:
                partial_coords.append((stop_coords[-1], seq_end))
        else:
            if (stop_coords[0] - seq_start) >= min_len:
                partial_coords.append((seq_start, stop_coords[0]))

    # Captures partial 3' pORFs lacking a stop codon in the current reading frame
    elif start_coords.size != stop_coords.size == 0:
        if rf > 3:
            tmp_start = sorted(start_coords[start_coords >= min_len])
        else:
            tmp_start = sorted(start_coords[abs(seq_len - start_coords) >= min_len])
        if tmp_start:
            partial_coords += check_start_dist(tmp_start, [seq_start, seq_end], rf)

    else:
        # Keep first and last stop codons to check for partial ORF ends
        eval_stops = sorted([min(stop_coords), max(stop_coords)])

        if rf > 3:
            # Capture partial 5' ORFs
            if (seq_end - eval_stops[1]) >= min_len:
                if eval_stops[1] > max(start_coords):
                    partial_coords.append((eval_stops[1], seq_end))
                elif abs(eval_stops[1] - max(start_coords)) < min_len:
                    partial_coords.append((eval_stops[1], seq_end))

            # Capture partial 3' ORFs
            eval_start = [i for i in start_coords if i <= eval_stops[0] and abs(i - seq_start) >= min_len]
            if eval_start:
                partial_coords += check_start_dist(eval_start, [seq_start, seq_end], rf)

        else:
            # Capture partial 5' ORFs
            if (eval_stops[0] - seq_start) >= min_len:
                if eval_stops[0] < min(start_coords):
                    partial_coords.append((seq_start, eval_stops[0]))
                elif abs(eval_stops[0] - min(start_coords)) < min_len:
                    partial_coords.append((seq_start, eval_stops[0]))

            # Capture partial 3' ORFs
            eval_start = [i for i in start_coords if i >= eval_stops[-1] and abs(i - seq_end) >= min_len]
            if eval_start:
                partial_coords += check_start_dist(eval_start, [seq_start, seq_end], rf)

    return partial_coords


def eval_pORF_coords(start_coords: np.ndarray, stop_coords: np.ndarray, rf: int, min_len: int, seq_len: int):
    """
    Evaluates combinations of start and stop coordinates to return a list of
    putative ORF (pORF) coordinates from a given transcript.

    Parameters
    ----------
    start_coords:  numpy array of start codon positions
    stop_coords:   numpy array of stop codon positions
    rf:            current reading frame
    min_len:       minimum ORF length to consider [default is 300nt]

    Returns
    ----------
    porf_coords:  list of pORF coordinates in the given reading frame
    """

    porf_coords = []
    valid_stops = []

    if rf > 3:
        seq_start = rf - 4
        seq_end = seq_len - (seq_len - seq_start)%3
        if stop_coords.size > 0 and abs(stop_coords[-1] - seq_end) >= min_len:
            valid_stops.append((stop_coords[-1], seq_end))

    else:
        seq_start = rf - 1
        if stop_coords.size  > 0 and abs(stop_coords[0] - seq_start) >= min_len:
            valid_stops.append((seq_start, stop_coords[0]))

    # keep windows between stop codons that are at least "min_len" nucleotides apart
    valid_stops += [(j, stop_coords[i+1]) for i, j in enumerate(stop_coords[:-1]) if abs(j - stop_coords[i+1]) >= min_len]

    if valid_stops:
        for stop_coord in valid_stops:
            if rf > 3:
                # Captures all start codon positions between stop codons that would
                # create ORFs larger than the minimum ORF size
                tmp_start = sorted(
                    [i for i in start_coords[start_coords <= stop_coord[1]] if (i - stop_coord[0]) >= min_len],
                    reverse = True
                    )
            else:
                # captures all start codon positions between stop codons that would
                # create ORFs larger than the minimum ORF size
                tmp_start = sorted(
                    [i for i in start_coords[start_coords >= stop_coord[0]] if (stop_coord[1] - i) >= min_len]
                    )
            if tmp_start:
                porf_coords += check_start_dist(tmp_start, stop_coord, rf)

    porf_coords += check_partial_pORFs(start_coords, stop_coords, rf, min_len, seq_len)

    return porf_coords


def prepare_pORF_seqs(seq_num: int, seq_name: str, seq: str, coord: list, fstop: list, rf: int):
    """
    Prepares the pORFs from a set of transcript coordinates.

    Parameters
    ----------
    seq_num:   the current pORF number to append to name
    seq_name:  original transcript name
    seq:       the transcript nucleotide sequence
    coord:     ORF start-stop coordinates
    fstop:     all stop codons on the plus strand/orienatation
    rf:        current reading frame

    Returns
    ----------
    porf:            ORF nucleotide sequence
    porf_name:       name of the putative ORF
    porf_full_name:  the detailed pORF name (include completeness, length, coordinates)
    """

    porf = seq[min(coord):max(coord)]
    porf_len = f'orf_length:{len(porf)}'
    porf_name = f'{seq_name}.pORF{seq_num}'

    if rf > 3:
        porf = f'{Seq(porf).reverse_complement()}'

        if porf[:3] == 'ATG' and porf[-3:] in fstop:
            comp = 'orf_type:complete'

        elif porf[:3] != 'ATG' and porf[-3:] in fstop:
            comp = 'orf_type:5prime_partial'

        else:
            comp = 'orf_type:3prime_partial'

        porf_full_name = f'{porf_name} {comp} {porf_len} {seq_name}:{max(coord)}-{min(coord)+1}(-)'

    else:
        if porf[:3] == 'ATG' and porf[-3:] in fstop:
            comp = 'orf_type:complete'

        elif porf[:3] != 'ATG' and porf[-3:] in fstop:
            comp = 'orf_type:5prime_partial'

        elif porf[:3] == 'ATG' and porf[-3:] not in fstop:
            comp = 'orf_type:3prime_partial'

        else:
            comp = 'orf_type:incomplete'

        porf_full_name = f'{porf_name} {comp} {porf_len} {seq_name}:{coord[0]+1}-{coord[1]}(+)'

    return porf_name, porf_full_name, porf


def extract_orfs_from_seq(seq_name: str, seq: str, fstop, rstop, min_len: int, stranded: str = None):
    """
    Extracts all the putative ORFs greater than some minimum length for all reading
    frames (unless stranded RNA-seq/assembly) for a given transcript.

    Parameters
    ----------
    query:     name of the "filtered" transcriptome's FASTA formatted file
    seq:       the transcript nucleotide sequence
    fstop:     all stop codons on the plus strand
    rstop:     all stop codons on the minus strand (rev. complement of fstop)
    min_len:   minimum ORF length to consider [default is 300nt]
    stranded:  designates from which strand (if not both) to call putative ORFs

    Returns dictionary of all putative ORFs (partials included)
    """

    all_pORFs = {}

    start_rf, stop_rf = capture_start_stop_pos(seq, fstop, rstop)

    # Remove reading frames that shouldn't be considered due to
    # stranded-sequencing and assembly protocols
    if stranded and stranded.lower() == 'plus':
        del stop_rf[4]
        del stop_rf[5]
        del stop_rf[6]

    elif stranded and stranded.lower() == 'minus':
        del stop_rf[1]
        del stop_rf[2]
        del stop_rf[3]

    else:
        pass

    porf_num = 1

    for k in stop_rf.keys():
        stop_coords = np.array(sorted(stop_rf[k]))
        start_coords = np.array(sorted(start_rf[k]))

        porf_coords = eval_pORF_coords(
                        start_coords,
                        stop_coords,
                        k,
                        min_len,
                        len(seq))
        # If a given reading frame has pORFs, then prepare them...
        if porf_coords:
            for coord in porf_coords:
                prepped_porf = prepare_pORF_seqs(porf_num, seq_name, seq, coord, fstop, k)

                all_pORFs[prepped_porf[0]] = prepped_porf[1:]

                porf_num += 1

    return all_pORFs


def ref_orf_dmnd(query: str, taxon_code: str, dmnd_db: str, min_len: int, evalue: float, threads: int, intermed: bool) -> list:
    """
    Runs DIAMOND BLASTX against a reference protein database.

    Parameters
    ----------
    query:       name of the "filtered" transcriptome's FASTA formatted file
    taxon_code:  species/taxon name or abbreviated code
    dmnd_db:     path to a protein database for DIAMOND
    min_len:     minimum ORF length to consider
    evalue:      maximum e-value to consider hits from DIAMOND
    threads:     number of threads to use
    intermed:    keep intermediate TSV outputs from DIAMOND

    Returns list of "BLASTX" hits for parsing
    """

    outfmt = '6 qseqid sseqid pident length qstart qend sstart send evalue bitscore qframe'

    if intermed:
        tsv_out = f'{taxon_code}_TIdeS/ORF_Calling/{taxon_code}.{min_len}bp.DMND_REF_ORFs.tsv'
        outfmt = f'{outfmt} -o {tsv_out}'

    dmnd_cmd = f'diamond blastx ' \
                f'-p {threads} ' \
                f'-e {evalue} ' \
                f'-q {query} ' \
                f'-d {dmnd_db} ' \
                f'-k 1 ' \
                f'-f {outfmt}'

    dmnd_rslt = subprocess.run(dmnd_cmd.split(),
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    if intermed:
        return [i.rstrip() for i in open(outfmt.split()[-1]).readlines()]

    else:
        return dmnd_rslt.stdout.split('\n')[:-1]


def generate_ref_orfs(fasta_file: str, taxon_code: str, start_time, dmnd_db: str, min_len: int, evalue: float, threads: int, verb: bool) -> dict:
    """
    Generate reference ORFs from outputs of DIAMOND BLASTX against a reference
    protein database.

    Parameters
    ----------
    query:       name of the "filtered" transcriptome's FASTA formatted file
    fasta_file:  name of the "filtered" transcriptome's FASTA formatted file
    gcode:       NCBI numerical genetic code
    rstop:       all stop codons on the minus strand (rev. complement of fstop)
    min_len:     minimum ORF length to consider [default is 300nt]
    stranded:    designates from which strand (if not both) to call putative ORFs

    Returns
    ----------
    ref_orf_db: dictionary of reference ORFs
    """

    ref_orf_db = {}
    ref_orfs = []

    if verb:
        print(f'[{timedelta(seconds=round(time.time()-start_time))}]  Running DIAMOND BLASTX against protein database')

    dmnd_hits = {i.split('\t')[0]:i.split('\t') for i in
                    ref_orf_dmnd(fasta_file, taxon_code, dmnd_db, min_len, evalue, threads, True)}

    if verb:
        print(f'[{timedelta(seconds=round(time.time()-start_time))}]  Generating training ORFs for {taxon_code}')

    for seq_rec in SeqIO.parse(fasta_file,'fasta'):
        if seq_rec.id in dmnd_hits:
            orf_pos = [int(j) for j in dmnd_hits[seq_rec.id][4:6]]
            orf = f'{seq_rec.seq[min(orf_pos)-1:max(orf_pos)]}'

            if orf_pos[0] > orf_pos[1]:
                ref_orf_db[seq_rec.id] = f'{Seq(orf).reverse_complement()}'
                ref_orfs.append(SeqRecord(Seq(orf).reverse_complement(),seq_rec.id,'',''))

            else:
                ref_orf_db[seq_rec.id] = orf
                ref_orfs.append(SeqRecord(Seq(orf),seq_rec.id,'',''))

    return ref_orf_db


def capture_pORFs(fasta_file: str, taxon_code: str, start_time, gcode: int = 1, min_len: int = 300, partial: bool = False, stranded: str = None, verb: bool = True):
    """
    Manages the capture of all putative ORFs from all transcripts in a given
    FASTA file.

    Parameters
    ----------
    fasta_file:  name of the "filtered" transcriptome's FASTA formatted file
    taxon_code:  species/taxon name or abbreviated code
    start_time:  initial timestamp to track runtime
    gcode:       NCBI numerical genetic code
    min_len:     minimum ORF length to consider [default is 300nt]
    stranded:    designates from which strand (if not both) to call putative ORFs
    verb:        verbose print statements

    Returns
    ----------
    pORF_db:     dictionary of all putative ORFs (partials included)
    porf_fasta:  FASTA formatted file of all putative ORFs
    """

    if verb:
        print(f'[{timedelta(seconds=round(time.time()-start_time))}]  Extracting putative ORFs for {taxon_code}')

    porf_dir = f'{taxon_code}_TIdeS/ORF_Calling/'
    porf_fasta = f'{porf_dir}{taxon_code}.{min_len}bp.pORFs.fas'

    Path(porf_dir).mkdir(exist_ok = True, parents = True)

    gcode_info = gcode_start_stops(gcode)

    pORF_db = {}
    all_pORFs = []

    for i in SeqIO.parse(fasta_file, 'fasta'):
        seq_porfs = extract_orfs_from_seq(
                        i.id,
                        f'{i.seq}',
                        gcode_info[0],
                        gcode_info[1],
                        min_len,
                        stranded)
        if seq_porfs:
            for i in seq_porfs.values():
                if not partial and 'orf_type:complete' not in i[0]:
                    continue
                all_pORFs.append(SeqRecord(Seq(i[1]), i[0].split()[0], '', i[0]))

                pORF_db[i[0]] = i[1]

    SeqIO.write(all_pORFs, porf_fasta, 'fasta')

    return pORF_db, porf_fasta



#!/usr/bin/env python3

"""
Prepares query and training ORFs for subsequent classification.

Dependencies include: BioPython, Scikit-learn.
"""

import os, random, sys

import numpy as np

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from sklearn.feature_extraction.text import CountVectorizer


def trim_seq(seq: str, rf: int) -> str:
    """
    Trims sequences to ensure multiple of 3.

    Parameters
    ----------
    seq:  nucleotide sequence to trim
    rf:   sequence's reading frame

    Returns sequence with length divisble by 3.
    """

    temp = seq[rf-1:]

    return temp[:len(temp)-len(temp)%3]


def randomize_orientation(ref_orf_db: dict, taxon_code: str) -> dict:
    """
    Randomly changes the orientation in a subset of the training ORFs.

    Parameters
    ----------
    ref_orf_db:  dictionary of all the training ORFs
    taxon_code:  species/taxon name or abbreviated code

    Returns
    ----------
    rnd_ornt:  dictionary of all training ORFs and their classification (0/1)

    """

    rnd_ornt = {}
    rnd_ornt_seqs = []

    # Shuffle the dictionary keys
    seq_subsamp = random.sample(list(ref_orf_db.keys()), int(len(ref_orf_db)))


    for n in range(0, len(seq_subsamp)):
        ref_seq_name = f'{seq_subsamp[n]}_RF1'
        ref_seq = ref_orf_db[seq_subsamp[n]]


        rf = random.sample([-3,-2,-1, 2, 3], 1)[0]
        if rf > 0:
            rnd_seq_name = f'{seq_subsamp[n]}_RF{rf}'
            rnd_seq = trim_seq(ref_orf_db[seq_subsamp[n]], rf)

        else:
            rnd_seq_name = f'{seq_subsamp[n]}_RF{abs(rf) + 3}'
            rnd_seq = trim_seq(f'{Seq(ref_orf_db[seq_subsamp[n]]).reverse_complement()}', abs(rf))

        rnd_ornt[ref_seq_name] = [1, ref_seq]
        rnd_ornt[rnd_seq_name] = [0, rnd_seq]

        rnd_ornt_seqs.append(SeqRecord(Seq(ref_seq), ref_seq_name, '',''))
        rnd_ornt_seqs.append(SeqRecord(Seq(rnd_seq), rnd_seq_name, '',''))

        # Keep 50% of the training ORFs in the correct orientation
        # if n < len(seq_subsamp) * 0.5:
        #     rnd_seq_name = f'{seq_subsamp[n]}_RF1'
        #     rnd_seq = ref_orf_db[seq_subsamp[n]]
        #     rnd_ornt[rnd_seq_name] = [1, rnd_seq]
        #
        # else:
        #     # Randomly select "incorrect" reading frames
        #     rf = random.sample([-3,-2,-1, 2, 3], 1)[0]
        #
        #     if rf > 0:
        #         rnd_seq_name = f'{seq_subsamp[n]}_RF{rf}'
        #         rnd_seq = trim_seq(ref_orf_db[seq_subsamp[n]], rf)
        #
        #     else:
        #         rnd_seq_name = f'{seq_subsamp[n]}_RF{abs(rf) + 3}'
        #         rnd_seq = trim_seq(f'{Seq(ref_orf_db[seq_subsamp[n]]).reverse_complement()}', abs(rf))
        #
        #     rnd_ornt[rnd_seq_name] = [0, rnd_seq]
        #
        #     rnd_ornt_seqs.append(SeqRecord(Seq(rnd_seq), rnd_seq_name, '',''))

    train_fas = f'{taxon_code}_TIdeS/ORF_Calling/{taxon_code}.TrainingORFs.fas'

    SeqIO.write(rnd_ornt_seqs, train_fas, 'fasta')

    return rnd_ornt


def chunk_seq(seq: str, contam: bool, overlap: bool, kmer: int, step: int):
    """
    Chunks the sequences into appropriate "words" by kmer-length and with/out
    overlapping kmers with a given step size.

    Parameters
    ----------
    seq:      nucleotide sequence to trim
    contam:   training ORFs are for contamination calling
    overlap:  overlapping kmers permitted if True
    kmer:     kmer size (number of nt) to convert sequences into
    step:     nucleotide distance to consider for overlapping kmers

    Returns kmers of the given sequence with/out overlapping (and appropriate steps)
    """

    if overlap:
        if not step:
            step = int(kmer/2)
        kmer_list = [seq[n:n+kmer] for n in range(0, len(seq)-(kmer-1), step)]

    else:
        kmer_list = [seq[n:n+kmer] for n in range(0, len(seq), kmer)]

    return ' '.join(kmer_list)


def kmer_ngram_counts(train_orfs_dict: dict, query_orfs_dict: dict, taxon_code: str, partial: bool = False, cvec = None, contam: bool = False, overlap: bool = False, kmer: int = 3, step = None):
    """
    Generates the training and query arrays for subsequent classification.

    Parameters
    ----------
    train_orfs_dict:  dictionary of training ORFs
    query_orfs_dict:  dictionary of all putative ORFs (partials included)
    taxon_code:       species/taxon name or abbreviated code
    partial:          evaluate partial ORFs
    cvec:             pretrained CountVectorizer from prior run
    contam:           training ORFs are for contamination calling
    overlap:          overlapping kmers permitted if True
    kmer:             kmer size (number of nt) to convert sequences into
    step:             nucleotide distance to consider for overlapping kmers

    Returns
    ----------
    train_labels:  sequence names for the training dataset
    X_train_orfs:  classification-ready data for prepared training ORFs
    train_class:   labels (0/1) for the training dataset (X_train_orfs)
    query_labels:  sequence names for the query pORF dataset
    X_query_orfs:  classification-ready data for prepared query pORFs
    cvec:          trained CountVectorizer
    """

    train_labels, train_seqs, train_class = [],[],[]
    query_labels, query_seqs = [],[]

    for k, v in query_orfs_dict.items():
        if not contam:
            if not partial and 'orf_type:complete' in k:
                    query_seqs.append(chunk_seq(v, contam, overlap, kmer, step))
                    query_labels.append(k)
            else:
                query_seqs.append(chunk_seq(v, contam, overlap, kmer, step))
                query_labels.append(k)
        else:
            query_seqs.append(chunk_seq(v, contam, overlap, kmer, step))
            query_labels.append(k)

    if cvec:
        X_query_orfs = np.array([i/sum(i) for i in cvec.transform(query_seqs).toarray()])
        return None, (query_labels, X_query_orfs), cvec

    if not contam:
        train_orfs_dict = randomize_orientation(train_orfs_dict, taxon_code)

    for k, v in train_orfs_dict.items():
        train_seqs.append(chunk_seq(v[1], contam, overlap, kmer, step))
        train_labels.append(k)
        train_class.append(v[0])

    cvec = CountVectorizer(ngram_range = (1,1))

    X_query_orfs = np.array([i/sum(i) for i in cvec.fit_transform(query_seqs).toarray()])
    X_train_orfs = np.array([i/sum(i) for i in cvec.transform(train_seqs).toarray()])

    return (train_labels, X_train_orfs, train_class), (query_labels, X_query_orfs), cvec



#!/usr/bin/env python3

"""
Something something...


Dependencies include: Optuna, Scikit-learn, and XGBoost.
"""

import pickle, sys

from collections import defaultdict

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord


"""Need to add a log file with params, so that if given those data, then all options
are set to be the same...

This can be stored in the TIdeS.pkl file too..."""


def save_model(taxon_code, overlap, kmer, step, cvec, clf):
    mdl_pkl = f'{taxon_code}_TIdeS/{taxon_code}.TIdeS.pkl'

    with open(mdl_pkl, 'wb') as fout:
        pickle.dump((overlap, kmer, step, cvec, clf), fout)

def save_seqs(taxon_code, all_orfs, clf_summary, gcode, single_best = True, contam = False):
    if not contam:
        ntd_seqs = []
        aa_seqs = []
        if single_best:
            out_ntd = f'{taxon_code}_TIdeS/{taxon_code}.TIdeS.fasta'
            for k, v in clf_summary[1].items():
                s = v[0][1]

                ntd_seqs.append(SeqRecord(Seq(all_orfs[s]), s.split()[0], '', s))

                aa_seqs.append(SeqRecord(Seq(all_orfs[s]).translate(table = gcode), s.split()[0], '', s))

        else:
            out_ntd = f'{taxon_code}_TIdeS/{taxon_code}.TIdeS.All_CRF.fasta'

        out_aa = out_ntd.replace("TIdeS.","TIdeS.AA.")

        SeqIO.write(ntd_seqs, out_ntd, 'fasta')
        SeqIO.write(aa_seqs, out_aa, 'fasta')

    else:
        class_seqs = defaultdict(list)

        for k, v in all_orfs.items():
            class_seqs[clf_summary[k][0]].append(SeqRecord(Seq(v), k, '', ''))

        for k, v in class_seqs.items():
            out_ntd =  f'{taxon_code}_TIdeS/{taxon_code}.TIdeS.{k}.fasta'
            SeqIO.write(v, out_ntd, 'fasta')
        return class_seqs.keys()



from setuptools import setup
from setuptools import find_packages

with open("README.md", 'r') as f:
    long_description = f.read()

setup(
   python_requires = ">=3.8.0" and "<= 3.11.7",
   name='TIdeS-ML',
   version='1.2.0',
   description='Tool for ORF-calling and ORF-classification using ML approaches',
   license="MIT",
   long_description = long_description,
   long_description_content_type='text/markdown',
   author='Xyrus X. Maurer-Alcala',
   author_email='xmaurer-alcala@amnh.org',
   url="https://github.com/xxmalcala/TIdeS",
   # package_dir = {'':'tides'},
   packages = ['tides','tides/bin'],
   # packages = find_packages('tides'),
   install_requires=['biopython>=1.79', 'ete3>=3.1.2', 'optuna>=3.0', 'pandas>=2.0', 'seaborn>=0.12.2', 'scikit-learn>=1.3.0'],
   entry_points = {
        'console_scripts': ['tides = tides.tides:main'],
    },
   classifiers=[
       "Intended Audience :: Science/Research",
       "Intended Audience :: Developers",
       "Intended Audience :: Science/Research",
       "License :: Freely Distributable",
       "Development Status :: 5 - Production/Stable",
       "Operating System :: OS Independent",
       "Programming Language :: Python",
       "Programming Language :: Python :: 3",
       "Programming Language :: Python :: 3.8",
       "Programming Language :: Python :: 3.9",
       "Programming Language :: Python :: 3.10",
       "Programming Language :: Python :: 3.11",
       "Topic :: Scientific/Engineering",
       "Topic :: Scientific/Engineering :: Bio-Informatics"
    ],
)



#!/usr/bin/env python3

"""
Transcript Identification and Selection (TIdeS) is a too for bespoke ORF calling
and sequence classification based on limited information and classic machine learning
approaches.

Overall, the intent was to aid in ORF-calling and decontamination of single-cell
transcriptomes from largely uncultivable microeukaryotes, as the popular tool
TransDecoder underperforms in these taxa.

TIdeS does learn best from quality data defined by the user, and will generate
trained models to speed up ORF-calling and ORF-classification from similar taxa
or replicates of the same taxon.
"""

import argparse, glob, logging, os, pickle, shutil, sys, time
from datetime import datetime, timedelta
from pathlib import Path

from tides.bin import filt_seqs as ft
from tides.bin import orf_call as oc
from tides.bin import orf_prep as op
from tides.bin import classify_orfs as co
from tides.bin import save_preds as sp


def collect_args():

    parser = argparse.ArgumentParser(description = f'{ascii_logo_vsn()}\n{usage_msg()}',
            usage=argparse.SUPPRESS, add_help = False,
            formatter_class=argparse.RawDescriptionHelpFormatter)

    g = parser.add_argument_group('General Options', description = (
    '''--fin (-i)            input file in FASTA format\n'''
    '''--taxon (-o)          taxon or output name\n'''
    '''--threads (-t)        number of CPU threads (default = 1)\n'''
    '''--kraken (-k)         kraken2 database to identify and filter\n'''
    '''                      non-eukaryotic sequences\n'''
    '''--no-filter           skip the rRNA and transcript clustering steps\n'''
    '''--model (-m)          previously trained TIdeS model (".pkl" file)\n'''
    '''--kmer                kmer size for generating sequence features (default = 3)\n'''
    '''--overlap             permit overlapping kmers (see --kmer)\n'''
    '''--step                step-size for overlapping kmers\n'''
    '''                      (default is kmer-length/2)\n'''
    '''--clean               remove intermediate filter-step files\n'''
    '''--quiet (-q)          no console output\n'''
    '''--gzip (-gz)          tar and gzip TIdeS output\n'''
    '''--help (-h)           show this help message and exit'''))

    g.add_argument('--help', '-h', action="help", help = argparse.SUPPRESS)

    g.add_argument('--fin', '-i', action = 'store',
        metavar = ('[FASTA File]'), type = str,
        help = argparse.SUPPRESS)

    g.add_argument('--taxon','-o', nargs='+',
        action = 'store', metavar = '[Taxon]', type = str,
        help = argparse.SUPPRESS)

    g.add_argument('--threads','-t', action = 'store',
        default = 1, metavar = '[Threads]', type = int,
        help = argparse.SUPPRESS)

    g.add_argument('--kraken','-k', action = 'store',
        metavar = '[Kraken2 Database]', type = str, default = None,
        help = argparse.SUPPRESS)

    g.add_argument('--no-filter', action = 'store_true',
        help = argparse.SUPPRESS)

    g.add_argument('--model','-m', action = 'store',
        metavar = '[Trained-RFC]', type = str, default = None,
        help = argparse.SUPPRESS)

    g.add_argument('--kmer', action = 'store',
        default = 3, metavar = '[kmer]', type = int,
        help = argparse.SUPPRESS)

    g.add_argument('--overlap', action = 'store_true',
        help = argparse.SUPPRESS)

    g.add_argument('--step', action = 'store',
        type = int, default = None,
        help = argparse.SUPPRESS)

    g.add_argument('--clean', action = 'store_true',
        help = argparse.SUPPRESS)

    g.add_argument('--quiet','-q', action = 'store_true',
        help = argparse.SUPPRESS)

    g.add_argument('--gzip','-gz', action = 'store_true',
        help = argparse.SUPPRESS)

    g.add_argument('--version', action = 'store_true',
        help = argparse.SUPPRESS)

    porf = parser.add_argument_group('ORF-Calling Options', description = (
    '''--db (-d)             protein database (FASTA or DIAMOND format)\n'''
    '''--partial (-p)        evaluate partial ORFs as well\n'''
    '''--id (-id)            minimum % identity to remove redundant transcripts\n'''
    '''                      (default = 97)\n'''
    '''--memory              memory limit (MB) for CD-HIT (default = 2000, unlimited = 0)\n'''
    '''--min-orf (-l)        minimum ORF length (bp) to evaluate (default = 300)\n'''
    '''--max-orf (-ml)       maximum ORF length (bp) to evaluate (default = 10000)\n'''
    '''--evalue (-e)         maximum e-value to infer reference ORFs\n'''
    '''                      (default = 1e-30)\n'''
    '''--gencode (-g)        genetic code to use to translate ORFs\n'''
    '''--strand (-s)         query strands to call ORFs\n'''
    '''                      (both/minus/plus, default = both)'''))

    porf.add_argument('--db', '-d', action = 'store',
        metavar = '[Protein Database]', type = str,
        help = argparse.SUPPRESS)

    porf.add_argument('--partial','-p', action = 'store_true',
        help = argparse.SUPPRESS)

    porf.add_argument('--pid', '--id', '-id', action = 'store',
        default = 97, metavar = '[perc-identity]', type = int,
        help = argparse.SUPPRESS)

    porf.add_argument('--memory', action = 'store',
        default = 2000, metavar = '[memory-limit]', type = int,
        help = argparse.SUPPRESS)

    porf.add_argument('--min-orf', '-l', action = 'store',
        default = 300, metavar = '[bp]', type = int,
        help = argparse.SUPPRESS)

    porf.add_argument('--max-orf', '-ml', action = 'store',
        default = 10000, metavar = '[bp]', type = int,
        help = argparse.SUPPRESS)

    porf.add_argument('--evalue', '-e', action = 'store',
        default = 1e-30, metavar = '[e-value]', type = float,

        help = argparse.SUPPRESS)

    porf.add_argument('--gencode', '-g', action = 'store',
        default = '1', metavar = '[Genetic-Code]', type = str,
        help = argparse.SUPPRESS)

    porf.add_argument('--strand', '-s', choices = ['both', 'plus', 'minus'], action = 'store',
        default = 'both', type = str, help = argparse.SUPPRESS)

    cntm = parser.add_argument_group('Contamination-Calling Options', description = (
    '''--contam (-c)         table of annotated sequences\n'''))

    cntm.add_argument('--contam','-c', nargs = '?', const = True,
        help = argparse.SUPPRESS)

    # Ensures that just script description provided if no arguments provided
    if len(sys.argv[1:]) == 0:
        print(ascii_logo_vsn())
        print(detail_msg())
        print(f'{usage_msg()}\n')
        sys.exit()

    args = parser.parse_args()

    if args.version:
        print(ascii_logo_vsn())
        sys.exit()

    elif not args.taxon or not args.fin:
        print('\nMissing required inputs: FASTA-File and Output Name')
        print(f'\n{usage_msg()}\n')
        sys.exit()

    args.taxon = '_'.join(args.taxon)

    return args


def ascii_logo_vsn():
    alv_msg = """      _____ ___    _     ___
     |_   _|_ _|__| |___/ __|
       | |  | |/ _` / -_)__ \\
       |_| |___\__,_\___|___/

     Version 1.2.0
    """
    return alv_msg


def detail_msg():
    dtl_msg = """     TIdeS leverages supervised machine-learning to hunt
     and select "optimal" putative ORFs (pORFs) from a
     transcriptome and can infer ORFs from a target taxon
     from "noisy" (e.g. scRNA-seq) datasets.\n"""
    return dtl_msg


def usage_msg():
    return """Usage:\n    tides.py [options] --fin [FASTA file] --taxon [taxon name]"""


def check_dependencies(args):
    dpnds = [
        ('BioPython', 'Bio.SeqIO'), ('Optuna', 'optuna'), ('Pandas', 'pandas'),
        ('Scikit-Learn', 'sklearn.svm'), ('ete3', 'ete3'), ('Barrnap', 'barrnap'),
        ('CD-HIT', 'cd-hit-est'), ('DIAMOND', 'diamond'), ('Kraken2', 'kraken2')]

    dpnd_status = {}
    for n in range(9):
        if n < 5:
            try:
                __import__(dpnds[n][1])
                dpnd_status[dpnds[n][0]] = 'Check'
            except:
                dpnd_status[dpnds[n][0]] = 'Error'
        else:
            if shutil.which(dpnds[n][1]):
                dpnd_status[dpnds[n][0]] = 'Check'
            else:
                if not args.contam:
                    if not args.kraken:
                        if n == 8:
                            dpnd_status[dpnds[n][0]] = 'Warning'
                        else:
                            dpnd_status[dpnds[n][0]] = 'Error'
                else:
                    if n == 8 and args.kraken:
                        dpnd_status[dpnds[n][0]] = 'Error'
                    else:
                        dpnd_status[dpnds[n][0]] = 'Warning'

    if 'Error' in dpnd_status.values():
        print('Error: Status of all dependencies:')
        for k, v in dpnd_status.items():
            print(f'    {k}:  {v}')
        sys.exit()


def predict_orfs(
        fasta_file: str,
        taxon_code:str,
        dmnd_db: str,
        kraken_db: str,
        partial: bool,
        filter: bool,
        gcode: str,
        kmer: int,
        overlap: bool,
        step: int,
        model: str,
        min_len: int,
        max_len: int,
        pid: float,
        memory: int,
        evalue: float,
        threads: int,
        strand: str,
        verb: bool,
        ) -> None:

    """
    Predicts in-frame Open Reading Frames (ORFs) from a given transcriptome.

    Parameters
    ----------
    fasta_file:  FASTA formatted transcriptome file
    taxon_code:  species/taxon name or abbreviated code
    dmnd_db:     path to a protein database for DIAMOND
    partial:     evaluate partial ORFs
    gcode:       genetic code used for ORF-calling and translation steps
    kmer:        kmer size (number of nt) to convert sequences into
    overlap:     overlapping kmers permitted if True
    step:        step-size for overlapping kmers
    model:       pickle file from previous TIdeS run
    min_len:     minimum ORF length to consider
    max_len:     maximum ORF length to consider
    pid:         percent identity (0-1.0) for removing redundant sequences
    memory:      memory limit (MB) for CD-HIT
    evalue:      maximum e-value to keep hits from DIAMOND
    threads:     number of threads to use
    strand:      designate strand(s) for ORF calling
    verb:        verbose print statements

    Returns current time to track overall runtime.
    """

    sttime = time.time()
    ref_orfs = ref_orf_fas = cvec = clf = None
    stop_codons, rstop_codons, ttable = oc.gcode_start_stops(gcode)

    if model:
        with open(model, 'rb') as f:
            overlap, kmer, step, cvec, clf = pickle.load(f)

    elif overlap and not step:
        step = int(kmer/2)

    if verb:
        print('#------ Preparing Transcriptome Data -------#')

    logging.info(
        f'\n+---------------------------------+\n'\
        f'|      Transcript-Filtering       |\n'\
        f'+---------------------------------+'
        )

    filt_fas = ft.filter_transcripts(
                    fasta_file,
                    taxon_code,
                    sttime,
                    kraken_db,
                    filter,
                    min_len,
                    max_len,
                    threads,
                    pid,
                    memory,
                    verb
                    )

    orig_seqs = open(fasta_file).read().count('>')
    sfilt_seqs = open(filt_fas).read().count('>')

    logging.info(f'\nInitial Transcripts:   {orig_seqs}')
    logging.info(f'Filtered Transcripts:  {sfilt_seqs}')

    if verb:
        if not model:
            print('\n#------- Calling Training and pORFs --------#')
        else:
            print('\n#------------- Calling pORFs ---------------#')

    logging.info(
        f'\n+---------------------------------+\n'\
        f'|           ORF-Calling           |\n'\
        f'+---------------------------------+'
        )

    putative_orfs, porf_fas = oc.capture_pORFs(
                                    filt_fas,
                                    taxon_code,
                                    sttime,
                                    gcode,
                                    min_len,
                                    partial,
                                    strand,
                                    verb
                                    )

    porf_count = len(putative_orfs)
    logging.info(f'\nPutative ORFs Called:  {porf_count}')

    if not model:
        ref_orfs = oc.generate_ref_orfs(
                                    filt_fas,
                                    taxon_code,
                                    sttime,
                                    dmnd_db,
                                    min_len,
                                    evalue,
                                    threads,
                                    verb
                                    )

    if verb:
        if not model:
            print(f'[{timedelta(seconds=round(time.time()-sttime))}]  Preparing training and query ORFs for {taxon_code}')
        else:
            print(f'[{timedelta(seconds=round(time.time()-sttime))}]  Preparing query ORFs for {taxon_code}')

    train_data, query_data, cvec = op.kmer_ngram_counts(
                                    ref_orfs,
                                    putative_orfs,
                                    taxon_code,
                                    partial,
                                    cvec,
                                    False,
                                    overlap,
                                    kmer,
                                    None
                                    )

    if verb:
        print('\n#----------- ORF Classification ------------#')

    clf_summary, clf = co.classify_orfs(
                        taxon_code,
                        sttime,
                        train_data,
                        query_data,
                        threads,
                        clf,
                        False,
                        verb
                        )

    tds_orfs = len(clf_summary[1])
    logging.info(f'TIdeS ORF Predictions: {tds_orfs}')

    if verb:
        print('\n#---------- Saving TIdeS Outputs -----------#')
        print(f'[{timedelta(seconds=round(time.time()-sttime))}] '
                ' Making FASTA files and storing TIdeS model')

    sp.save_model(
        taxon_code,
        overlap,
        kmer,
        step,
        cvec,
        clf
        )

    sp.save_seqs(
        taxon_code,
        putative_orfs,
        clf_summary,
        ttable,
        True
        )

    logging.info(
        f'\n+---------------------------------+\n'\
        f'|          Main Outputs           |\n'\
        f'+---------------------------------+'
        )

    logging.info(f'\nPredicted ORFs:   {taxon_code}_TIdeS/{taxon_code}.TIdeS.fasta')
    logging.info(f'Trained Model:    {taxon_code}_TIdeS/{taxon_code}.TIdeS.pkl')
    logging.info(f'Classifications:  {taxon_code}_TIdeS/Classification/{taxon_code}.TIdeS_Classification.tsv')

    return sttime


def eval_contam(fasta_file: str,
                taxon_code: str,
                contam_list: str,
                kraken_db: list,
                gcode: str,
                kmer: int,
                overlap: bool,
                step: int,
                model: str,
                threads: int,
                verb: bool) -> None:

    """
    Binary classification of target versus non-target sequences

    Parameters
    ----------
    fasta_file: FASTA formatted transcriptome file
    taxon_code: species/taxon name or abbreviated code
    sister_summary: tab-delimited file with user-defined "target" and "non-target" sequences
    gcode: genetic code used for ORF-calling and translation steps
    pretrained: random forest model from previous TIdeS run
    min_len: minimum ORF length to consider
    threads: number of cpu threads to use
    verb: verbose print statements

    Returns current time to track overall runtime
    """


    sttime = time.time()
    cvec = clf = None
    stop_codons, rstop_codons, ttable = oc.gcode_start_stops(gcode)

    if model:
        with open(model, 'rb') as f:
            overlap, kmer, step, cvec, clf = pickle.load(f)

    elif overlap and not step:
        step = int(kmer/2)

    if verb:
        print('#---- Preparing User-Assessed ORF Data -----#')

    train_orfs, query_orfs = ft.prep_contam(
                                fasta_file,
                                taxon_code,
                                contam_list,
                                kraken_db,
                                model,
                                sttime,
                                threads,
                                verb
                                )

    if model:
        with open(model, 'rb') as f:
            overlap, kmer, step, cvec, clf = pickle.load(f)

    if verb:
        if not model:
            print(f'[{timedelta(seconds=round(time.time()-sttime))}]  Preparing training and query ORFs for {taxon_code}')
        else:
            print(f'[{timedelta(seconds=round(time.time()-sttime))}]  Preparing query ORFs for {taxon_code}')

    train_data, query_data, cvec = op.kmer_ngram_counts(
                                    train_orfs,
                                    query_orfs,
                                    taxon_code,
                                    False,
                                    cvec,
                                    True,
                                    overlap,
                                    kmer,
                                    step
                                    )


    if verb:
        print('\n#----------- ORF Classification ------------#')

    clf_summary, clf = co.classify_orfs(
                        taxon_code,
                        sttime,
                        train_data,
                        query_data,
                        threads,
                        clf,
                        True,
                        verb
                        )

    if verb:
        print('\n#---------- Saving TIdeS Outputs -----------#')
        print(f'[{timedelta(seconds=round(time.time()-sttime))}] '
                ' Making FASTA files and storing TIdeS model')

    sp.save_model(
        taxon_code,
        overlap,
        kmer,
        step,
        cvec,
        clf
        )

    orf_classes = sp.save_seqs(
                    taxon_code,
                    query_orfs,
                    clf_summary,
                    ttable,
                    True,
                    True
                    )
    logging.info(
        f'\n+---------------------------------+\n'\
        f'|     Classification Summary      |\n'\
        f'+---------------------------------+'
        )

    tds_cls_fas = []
    orig_seqs = open(fasta_file).read().count('>')

    logging.info(f'\nInitial ORFs:       {orig_seqs}')


    for k in list(orf_classes):
        cf = f'{taxon_code}_TIdeS/{taxon_code}.TIdeS.{k}.fasta'
        cf_cnt = open(cf).read().count('>')

        x = f'{k} ORFs:'.ljust(20)

        logging.info(f'{x}{cf_cnt}')

        tds_cls_fas.append(f'\n     {cf}')

    logging.info(
        f'\n+---------------------------------+\n'\
        f'|          Main Outputs           |\n'\
        f'+---------------------------------+'
        )

    logging.info(f'Trained Model:        {taxon_code}_TIdeS/{taxon_code}.TIdeS.pkl')
    logging.info(f'Classifications:      {taxon_code}_TIdeS/Classification/{taxon_code}.TIdeS_Classification.tsv')
    logging.info(f'Classified ORFs:{"".join(tds_cls_fas)}')

    return sttime


def init_log_stats(args):
    if args.model:
        with open(args.model, 'rb') as f:
            args.overlap, args.kmer, args.step, args.cvec, args.clf = pickle.load(f)

    arg_vars = vars(args)

    curr_time = datetime.now().strftime("%m/%d/%Y %I:%M %p")

    outdir = f'{arg_vars["taxon"]}_TIdeS/'

    Path(outdir).mkdir(parents = True, exist_ok = True)

    outlog = f'{outdir}/tides.log'

    logging.basicConfig(filename = outlog,
        filemode = 'w+',
        format = '%(message)s',
        level = logging.DEBUG)

    logging.info(
        f'+---------------------------------+\n'\
        f'|         TIdeS Parameters        |\n'\
        f'+---------------------------------+'
        )

    logging.info(f'\nRun-Start:  {curr_time}')

    if arg_vars['contam']:
        qparams = ['fin', 'taxon', 'contam', 'threads', 'kraken', 'model', 'kmer', 'overlap', 'step']
        logging.info(f'Mode:       ORF-Classification')

    else:
        qparams = ['fin', 'taxon', 'threads', 'kraken', 'no_filter', 'model', 'kmer',
                    'overlap', 'step', 'db', 'partial', 'pid', 'memory', 'min_orf',
                    'max_orf', 'evalue', 'gencode', 'strand']

        logging.info(f'\nMode:       ORF-Prediction')

    for k in qparams:
        x = f'{k}:'.ljust(12)
        logging.info(f'{x}{str(arg_vars[k] or "None")}')

    return outlog


def main():

    """Note to self:
    Add method to check the sister-summary table format...

    Requirements:
    -- Tab-delimited
    -- Target/Non-Target OR 1/0

    Also provide options for multi-classes (e.g., not solely current binary approach)...
    """

    args = collect_args()

    if not oc.gcode_start_stops(args.gencode):
        sys.exit(1)

    check_dependencies(args)

    init_log_stats(args)

    if float(args.pid) > 1:
        args.pid = args.pid / 100

    if not args.contam:
        if not args.db and not args.model:
            print('\nERROR: Missing diamond "blast" database -- use the "prep_tides_db.sh" '
                    'script in the Tides/util/ directory.\n')
            sys.exit(1)

        if not args.quiet:
            print(ascii_logo_vsn())

        sttime = predict_orfs(args.fin,
                                args.taxon,
                                args.db,
                                args.kraken,
                                args.partial,
                                args.no_filter,
                                args.gencode,
                                args.kmer,
                                args.overlap,
                                args.step,
                                args.model,
                                args.min_orf,
                                args.max_orf,
                                args.pid,
                                args.memory,
                                args.evalue,
                                args.threads,
                                args.strand,
                                not args.quiet)
    else:
        if not args.quiet:
            print(ascii_logo_vsn())

        if not isinstance(args.contam, str) and not args.kraken and not args.model:
            print(args)
            print('ERROR: Please include the path to a file of annotated sequences to classify OR\n' \
                'the path to a suitable Kraken2 database for automatic annotation of\n' \
                'non-eukaryotic sequences.\n')
            sys.exit(1)

        elif isinstance(args.contam, str) and not args.kraken and not Path(args.contam).is_file():
            print(f'Could not find the file {args.contam}. Please check that the correct\n' \
                'path is provided.\n')
            sys.exit(1)

        sttime = eval_contam(args.fin,
                            args.taxon,
                            args.contam,
                            args.kraken,
                            args.gencode,
                            args.kmer,
                            args.overlap,
                            args.step,
                            args.model,
                            args.threads,
                            not args.quiet)

    if args.clean:
        if not args.quiet:
            print(f'[{timedelta(seconds=round(time.time()-sttime))}]  Removing intermediate files')

        if Path(f'{args.taxon}_TIdeS/Filter_Steps/').exists():
            os.system(f'rm -rf {args.taxon}_TIdeS/Filter_Steps')
        if Path(f'{args.taxon}_TIdeS/ORF_Calling/').exists():
            os.system(f'rm -rf {args.taxon}_TIdeS/ORF_Calling/')

    if args.gzip:
        if not args.quiet:
            print(f'[{timedelta(seconds=round(time.time()-sttime))}]  Compressing TIdeS Outputs')

        os.system(f'tar -zcf {args.taxon}_TIdeS.tar.gz {args.taxon}_TIdeS/')

    if not args.quiet:
        print(f'[{timedelta(seconds=round(time.time()-sttime))}]  Finished running TIdeS')
        print('\n#--------- Thanks for using TIdeS! ---------#\n')



#!/usr/bin/env python3

"""
Prepares folder(s) and filters the transcriptome by length and putative
rRNA sequences.

Note that rRNA filtering is performed by Barrnap. Taxa with unusual rRNAs, such
as Foraminifera, will need to have their rRNAs removed independently.

Dependencies include: Barrnap, BioPython, CD-HIT.
"""

import numpy, random, shutil, subprocess, sys, time
from datetime import timedelta
from pathlib import Path

from Bio import SeqIO
from ete3 import NCBITaxa


def filt_len(fasta_file: str, taxon_code: str, out_dir: str, min_len: int, max_len: int) -> str:
    """
    Remove sequences shorter than a minimum length

    Parameters
    ----------
    fasta_file:  FASTA formatted file
    taxon_code:  species/taxon name or abbreviated code
    out_dir:     output directory to store filtered data
    min_len:     minimum ORF length to consider [default is 300nt]

    Returns
    ----------
    filt_len_fasta:  FASTA formatted file with sequences above a given minimum length
    """

    filt_len_fasta = f'{out_dir}/{taxon_code}.{min_len}bp.fas'
    filt_len_conv_tbl = f'{out_dir}/{taxon_code}.{min_len}bp.NameConversion.tsv'

    t_cnt = 0
    updated_seqs = []
    name_convert = []

    for i in SeqIO.parse(fasta_file, 'fasta'):
        if min_len <= len(i.seq) <= max_len:
            t_cnt += 1

            updated_seq_name = f'{taxon_code}|Transcript_{t_cnt}_Len_{len(i.seq)}'

            if 'cov' in i.id:
                cov_val = float(i.id.split("cov_")[1].split("_")[0])
                updated_seq_name += f'_Cov_{cov_val:.2f}'

            name_convert.append(f'{updated_seq_name}\t{i.description}\n')
            i.id = updated_seq_name
            i.description = ''
            i.seq = i.seq.upper()

            updated_seqs.append(i)

    SeqIO.write(updated_seqs, filt_len_fasta, 'fasta')

    with open(filt_len_conv_tbl,'w+') as w:
        w.write('Updated-Name\tOriginal-Name\n')
        w.write(''.join(name_convert))

    return filt_len_fasta


def run_barrnap(fasta_file: str, threads: int) -> list:
    """
    Run Barrnap to remove easily identifiable rRNAs.

    Parameters
    ----------
    fasta_file:  FASTA formatted file
    threads:     number of cpu threads to use

    Returns
    ----------
    rRNA_seqs:  list of putative rRNA sequence names
    """

    rRNA_seqs = []
    kdm = ['bac','arc','mito','euk']

    for k in kdm:
        bnp_cmd = ['barrnap', '--kingdom', k, '--threads', f'{threads}', fasta_file]

        bnp_rslt = subprocess.run(
                        bnp_cmd,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        universal_newlines=True
                        )

        rRNA_seqs += [i.split('\t')[0] for i in bnp_rslt.stdout.split('\n') if '|' in i]

    return rRNA_seqs


def remove_rRNA(fasta_file: str, taxon_code: str, out_dir: str, min_len: int, threads: int) -> str:
    """
    Remove putative rRNA sequences from FASTA file

    Parameters
    ----------
    fasta_file:  FASTA formatted file
    taxon_code:  species/taxon name or abbreviated code
    out_dir:     output directory to store filtered data
    min_len:     minimum ORF length to consider [default is 300nt]
    threads:     number of cpu threads to use

    Returns
    ----------
    rRNA_clean_fas:  FASTA formatted file without putative rRNA sequences
    """

    rRNA_seqs = run_barrnap(fasta_file, threads)

    rRNA_clean_fas = f'{out_dir}{taxon_code}.{min_len}bp.Filt_rRNA.fas'
    rRNA_fas = f'{out_dir}{taxon_code}.{min_len}bp.rRNA_Seqs.fas'

    rRNA_contam, clean_seqs = [],[]

    # bin sequences as either putative rRNAs or "clean"
    for i in SeqIO.parse(fasta_file, 'fasta'):
        if i.id in rRNA_seqs:
            rRNA_contam.append(i)
        else:
            clean_seqs.append(i)

    SeqIO.write(rRNA_contam, rRNA_fas, 'fasta')
    SeqIO.write(clean_seqs, rRNA_clean_fas, 'fasta')

    return rRNA_clean_fas


def run_kraken2(fasta_file: str, taxon_code: str, out_dir: str, kraken_db: str, threads: int) -> str:
    """
    Run Kraken2 to identify putative non-eukaryotic contamination

    Parameters
    ----------
    fasta_file:  FASTA formatted file
    taxon_code:  species/taxon name or abbreviated code
    out_dir:     output directory to store Kraken2's classification reports
    kraken_db:   path to kraken2-database directory
    threads:     number of cpu threads to use

    Returns
    ----------
    krak_out:   kraken2 taxonomic assignment report
    """

    krak_out = f'{out_dir}{taxon_code}.Kraken2_Txnmy.txt'

    krak_cmd = 'kraken2 --confidence 0.1 ' \
                f'--db {kraken_db} ' \
                f'--threads {threads} ' \
                f'--output {krak_out} ' \
                f'{fasta_file}'

    krak_rslt = subprocess.run(
                    krak_cmd.split(),
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    universal_newlines=True
                    )

    return krak_out


def assign_taxonomy(fasta_file: str, taxon_code: str, out_dir: str, kraken_db: str, threads: int) -> dict:
    """
    Runs Kraken2 and assigns taxonomy to the classified sequences.

    Parameters
    ----------
    fasta_file:  FASTA formatted file
    taxon_code:  species/taxon name or abbreviated code
    out_dir:     output directory to store Kraken2's classification reports
    kraken_db:   path to kraken2-database directory
    threads:     number of cpu threads to use

    Returns
    ----------
    krak_seq_txnmy: dictionary of sequences and their broad classification
    """

    kraken_output = run_kraken2(fasta_file, taxon_code, out_dir, kraken_db, threads)

    krak_seq_txnmy = {}

    ncbi = NCBITaxa()

    if not Path(f'{Path.home()}/.etetoolkit/taxa.sqlite').is_file():
        ncbi.update_taxonomy_database()

    with open(kraken_output, 'r') as fi:
        for ln in fi:
            if ln[0] == 'C':
                taxid = ln.split('\t')[2]
                txnmy = ncbi.get_taxid_translator(ncbi.get_lineage(taxid))
                if 'Bacteria' in txnmy.values():
                    krak_seq_txnmy[ln.split('\t')[1]] = 'Bacteria'
                elif 'Archaea' in txnmy.values():
                    krak_seq_txnmy[ln.split('\t')[1]] = 'Archaea'
                elif 'Viruses' in txnmy.values():
                    krak_seq_txnmy[ln.split('\t')[1]] = 'Virus'
                else:
                    krak_seq_txnmy[ln.split('\t')[1]] = 'PutativeEuk'

            else:
                krak_seq_txnmy[ln.split('\t')[1]] = 'PutativeEuk'

    return krak_seq_txnmy


def kraken_contam_eval(fasta_file: str, taxon_code: str, out_dir: str, kraken_db: str, threads: int) -> str:
    """
    Takes the taxonomic assignments to identify sequences for training TIdeS

    Parameters
    ----------
    fasta_file:  FASTA formatted file
    taxon_code:  species/taxon name or abbreviated code
    out_dir:     output directory to store Kraken2's classification reports and
                 sequence classification table
    kraken_db:   path to kraken2-database directory
    threads:     number of cpu threads to use

    Returns
    ----------
    seq_eval_summary: table of annotated sequences for training
    """


    seq_eval_summary = f'{out_dir}{taxon_code}.KrakenEval.TIdeS.txt'

    krak_seq_txnmy = assign_taxonomy(fasta_file, taxon_code, out_dir, kraken_db, threads)

    euk_seqs = []
    non_euk_seqs = []
    for k, v in krak_seq_txnmy.items():
        if v != 'PutativeEuk':
            non_euk_seqs.append(f'{k}\tNon-Eukaryotic\n')
        else:
            euk_seqs.append(f'{k}\t{taxon_code}\n')

    random.shuffle(non_euk_seqs)
    random.shuffle(euk_seqs)

    fin_data = non_euk_seqs[:min([len(non_euk_seqs), len(euk_seqs), 200])]
    fin_data += euk_seqs[:min([len(non_euk_seqs), len(euk_seqs), 200])]

    if not fin_data:
        print(f' ----- No sequences were annotated as "non-eukaryotic"')
        print(f' ----- Quitting TIdeS')
        sys.exit(1)

    with open(seq_eval_summary, 'w+') as w:
        w.write(''.join(fin_data))

    return seq_eval_summary


def remove_non_euk(fasta_file: str, taxon_code: str, out_dir: str, kraken_db: str, threads: int) -> str:
    """
    Remove putative non-eukaryotic contamination from FASTA file

    Parameters
    ----------
    fasta_file:  FASTA formatted file
    taxon_code:  species/taxon name or abbreviated code
    out_dir:     output directory to store annotated sequences (euk and non-euk)
    kraken_db:   path to kraken2-database directory
    threads:     number of cpu threads to use

    Returns
    ----------
    peuk_fasta': FASTA formatted file of putative eukaryotic transcripts
    """

    peuk_fasta = f'{out_dir}{taxon_code}.Kraken_Txnmy.PutativeEuk.fasta'

    krak_seq_txnmy = assign_taxonomy(fasta_file, taxon_code, out_dir, kraken_db, threads)

    non_euk_seqs = {'Bacteria':[],'Viruses':[], 'Archaea':[]}
    peuk_seqs = []

    for i in SeqIO.parse(fasta_file,'fasta'):
        if i.id not in krak_seq_txnmy:
            peuk_seqs.append(i)
        else:
            if krak_seq_txnmy[i.id] == 'PutativeEuk':
                peuk_seqs.append(i)
            else:
                non_euk_seqs[krak_seq_txnmy[i.id]].append(i)

    for k, v in non_euk_seqs.items():
        if v:
            outf = f'{out_dir}{taxon_code}.Kraken_Txnmy.{k}.fasta'
            SeqIO.write(v, outf, 'fasta')

    if not peuk_seqs:
        print(f'ERROR: No putative eukaryotic sequences are present for {taxon_code} from {fasta_file}')
        sys.exit(1)

    SeqIO.write(peuk_seqs, peuk_fasta, 'fasta')

    return peuk_fasta


def clust_txps(fasta_file: str, taxon_code: str, out_dir: str, pid: float, mem: int, threads: int) -> str:
    """
    Remove redundant sequences using a given minium percent identity

    Parameters
    ----------
    fasta_file:  FASTA formatted file
    taxon_code:  species/taxon name or abbreviated code
    out_dir:     output directory to store clustered transcripts
    pid:         minimum percent identity to cluster redundant sequences
    mem:         memory limit (MB) for cd-hit (default: 800, unlimited: 0)
    threads:     number of cpu threads to use

    Returns
    ----------
    clust_fas:  FASTA formatted file with redundant transcripts removed
    """

    clust_fas = f'{out_dir}{fasta_file.split("/")[-1].split("Filt")[0]}{int(pid*100)}pid.Cluster.fas'

    cd_hit_cmd = f'cd-hit-est -T {threads} '\
                    '-G 0 ' \
                    f'-c {pid} ' \
                    '-aS 1.0 ' \
                    '-aL .005 ' \
                    f'-M {int(mem)} ' \
                    f'-i {fasta_file} ' \
                    f'-o {clust_fas}'

    cdht_rslt = subprocess.run(
                    cd_hit_cmd.split(),
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    universal_newlines=True
                    )

    return clust_fas


def prep_contam(fasta_file: str, taxon_code: str, sis_smry: str, kraken_db: str, model: str, start_time, threads: int, verb: bool) -> dict:

    """
    Prepares target/non-target sequences from a FASTA formatted file using a user
    provided summary table.

    Parameters
    ----------
    fasta_file:  FASTA formatted file
    taxon_code:  species/taxon name or abbreviated code
    sis_smry:    tab-delimited file with user-defined "target" and "non-target" sequences
    pretrained:  random forest model from previous TIdeS run
    start_time:  initial timestamp to track runtime
    verb:        verbose print statements

    Returns
    ----------
    targ_seqs:      dictionary of exemplar target sequences considered to capture
    non_targ_seqs:  dictionary of exemplar non-target sequences to avoid
    """

    backup_dir = f'{taxon_code}_TIdeS/Original/'
    eval_seq_fas = f'{backup_dir}{taxon_code}.EvalSeqs.fas'
    query_orfs = {}

    if verb:
        print(f'[{timedelta(seconds=round(time.time()-start_time))}]  Backing up data for {taxon_code}')

    prep_dir(backup_dir)

    shutil.copy2(fasta_file, backup_dir)

    if not model:
        if kraken_db:
            if verb:
                print(f'[{timedelta(seconds=round(time.time()-start_time))}]  Classifying ORFs with Kraken2')

            sis_smry = kraken_contam_eval(fasta_file, taxon_code, backup_dir, kraken_db, threads)
        else:
            shutil.copy2(sis_smry, backup_dir)

    if verb:
        print(f'[{timedelta(seconds=round(time.time()-start_time))}]  Parsing ORF classifications')

    if not model:
        train_orfs = {i.split('\t')[0]:i.split('\t')[1].rstrip() for i in open(sis_smry).readlines()}
        contam_seqs = {i:0 for i in list(set(train_orfs.values()))}

    # Group sequences by user classification
    for i in SeqIO.parse(fasta_file, 'fasta'):
        if model:
            query_orfs[i.id] = f'{i.seq}'
        else:
            if i.id in train_orfs.keys():
                contam_seqs[train_orfs[i.id]] += 1
                train_orfs[i.id] = [train_orfs[i.id], f'{i.seq}']
            else:
                query_orfs[i.id] = f'{i.seq}'

    if model:
        return None, query_orfs

    if len(contam_seqs) < 2:
        print(f'[{timedelta(seconds=round(time.time()-start_time))}]  Error: ' \
            f'Missing at least TWO categories to run TIdeS')
        sys.exit(1)

    for k, v in contam_seqs.items():
        if v < 25:
            print(contam_seqs)
            print(f'[{timedelta(seconds=round(time.time()-start_time))}]  Error: ' \
                f'Fewer than 25 sequences were labeled as: {k}')

            print(f'[{timedelta(seconds=round(time.time()-start_time))}]  Exiting TIdeS: Too few training seqs.\n')
            sys.exit(1)

        elif 25 <= v < 100:
            print(f'[{timedelta(seconds=round(time.time()-start_time))}]  Warning: ' \
                f'Fewer than 100 sequences were labeled as: {k}')

        else:
            pass

    print(f'[{timedelta(seconds=round(time.time()-start_time))}]  Classification summary: ')
    for k, v in contam_seqs.items():
            print(f'{" "*11}----- {v} sequences were labeled as: {k}')

    return train_orfs, query_orfs


def prep_dir(new_dir: str) -> None:
    """
    Creates a new directory.

    Parameters
    ----------
    new_dir:  directory to create
    """

    Path(new_dir).mkdir(parents = True, exist_ok = True)


def filter_transcripts(
                        fasta_file: str,
                        taxon_code: str,
                        start_time,
                        kraken_db: str,
                        skip_filter: bool = False,
                        min_len: int = 300,
                        max_len: int = 10000,
                        threads: int = 4,
                        pid: float = 0.97,
                        mem: int = 2000,
                        verb: bool = True
                        ) -> str:
    """
    Performs all the initial filtering steps to ease classification

    Parameters
    ----------
    fasta_file:  FASTA formatted file
    taxon_code:  species/taxon name or abbreviated code
    start_time:  initial timestamp to track runtime
    min_len:     minimum ORF length to consider [default is 300nt]
    threads:     number of cpu threads to use
    pid:         minimum percent identity to cluster redundat sequences
    verb:        verbose print statements

    Returns
    ----------
    post_clst_fas:  FASTA formatted file of filtered transcripts for ORF-calling
    """

    backup_dir = f'{taxon_code}_TIdeS/Original/'
    filt_len_dir = f'{taxon_code}_TIdeS/Filter_Steps/Length_Filter/'
    clust_dir = f'{taxon_code}_TIdeS/Filter_Steps/Clustering/'
    krak_dir = f'{taxon_code}_TIdeS/Filter_Steps/Kraken_Filter/'
    filt_rRNA_dir = f'{taxon_code}_TIdeS/Filter_Steps/rRNA_Filter/'

    if verb:
        print(f'[{timedelta(seconds=round(time.time()-start_time))}]  Backing up data for {taxon_code}')

    prep_dir(backup_dir)
    shutil.copy2(fasta_file, backup_dir)

    if skip_filter:
        return f'{backup_dir}{fasta_file.rpartition("/")[-1]}'

    if verb:
        print(f'[{timedelta(seconds=round(time.time()-start_time))}]  Removing short transcripts')

    prep_dir(filt_len_dir)
    mlen_fas = filt_len(fasta_file, taxon_code, filt_len_dir, min_len, max_len)


    if verb:
        print(f'[{timedelta(seconds=round(time.time()-start_time))}]  Removing rRNA contamination')

    prep_dir(filt_rRNA_dir)
    rRNA_free = remove_rRNA(mlen_fas, taxon_code, filt_rRNA_dir, min_len, threads)

    if verb:
        print(f'[{timedelta(seconds=round(time.time()-start_time))}]  Filtering redundant transcripts')

    prep_dir(clust_dir)
    post_clst = clust_txps(rRNA_free, taxon_code, clust_dir, pid, mem, threads)

    if verb and kraken_db:
        print(f'[{timedelta(seconds=round(time.time()-start_time))}]  Removing non-eukaryotic transcripts')

        prep_dir(krak_dir)
        post_krak = remove_non_euk(post_clst, taxon_code, krak_dir, kraken_db, threads)

        return post_krak

    return post_clst
